@@ -9,6 +9,10 @@ header files (so that the JS compiler can see the constants in those
 headers, for the libc implementation in JS).
 '''
 
+from tools.toolchain_profiler import ToolchainProfiler
+if __name__ == '__main__':
+  ToolchainProfiler.record_process_start()
+
 import os, sys, json, optparse, subprocess, re, time, logging
 
 from tools import shared
@@ -647,7 +651,8 @@ function _emscripten_asm_const_%s(%s) {
              '(it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)' + \
              '"); ' + extra
 
-    basic_funcs = ['abort', 'assert'] + [m.replace('.', '_') for m in math_envs]
+    basic_funcs = ['abort', 'assert', 'enlargeMemory', 'getTotalMemory'] + [m.replace('.', '_') for m in math_envs]
+    if settings['ABORTING_MALLOC']: basic_funcs += ['abortOnCannotGrowMemory']
     if settings['STACK_OVERFLOW_CHECK']: basic_funcs += ['abortStackOverflow']
 
     asm_safe_heap = settings['SAFE_HEAP'] and not settings['SAFE_HEAP_LOG'] and not settings['RELOCATABLE'] # optimized safe heap in asm, when we can
@@ -663,11 +668,9 @@ function _emscripten_asm_const_%s(%s) {
         basic_funcs += ['nullFunc_' + sig]
         asm_setup += '\nfunction nullFunc_' + sig + '(x) { ' + get_function_pointer_error(sig) + 'abort(x) }\n'
 
-    basic_vars = ['STACKTOP', 'STACK_MAX', 'tempDoublePtr', 'ABORT']
+    basic_vars = ['STACKTOP', 'STACK_MAX', 'DYNAMICTOP_PTR', 'tempDoublePtr', 'ABORT']
     basic_float_vars = []
 
-    if settings['SAFE_HEAP']: basic_vars += ['DYNAMICTOP']
-
     if metadata.get('preciseI64MathUsed'):
       basic_vars += ['cttz_i8']
     else:
@@ -1059,7 +1062,7 @@ function setThrew(threw, value) {
 '''] + ['' if not settings['SAFE_HEAP'] else '''
 function setDynamicTop(value) {
   value = value | 0;
-  DYNAMICTOP = value;
+  HEAP32[DYNAMICTOP_PTR>>2] = value;
 }
 '''] + ['' if not asm_safe_heap else '''
 function SAFE_HEAP_STORE(dest, value, bytes) {
@@ -1067,7 +1070,7 @@ function SAFE_HEAP_STORE(dest, value, bytes) {
   value = value | 0;
   bytes = bytes | 0;
   if ((dest|0) <= 0) segfault();
-  if (((dest + bytes)|0) > (DYNAMICTOP|0)) segfault();
+  if (((dest + bytes)|0) > (HEAP32[DYNAMICTOP_PTR>>2]|0)) segfault();
   if ((bytes|0) == 4) {
     if ((dest&3)) alignfault();
     HEAP32[dest>>2] = value;
@@ -1083,7 +1086,7 @@ function SAFE_HEAP_STORE_D(dest, value, bytes) {
   value = +value;
   bytes = bytes | 0;
   if ((dest|0) <= 0) segfault();
-  if (((dest + bytes)|0) > (DYNAMICTOP|0)) segfault();
+  if (((dest + bytes)|0) > (HEAP32[DYNAMICTOP_PTR>>2]|0)) segfault();
   if ((bytes|0) == 8) {
     if ((dest&7)) alignfault();
     HEAPF64[dest>>3] = value;
@@ -1097,7 +1100,7 @@ function SAFE_HEAP_LOAD(dest, bytes, unsigned) {
   bytes = bytes | 0;
   unsigned = unsigned | 0;
   if ((dest|0) <= 0) segfault();
-  if ((dest + bytes|0) > (DYNAMICTOP|0)) segfault();
+  if ((dest + bytes|0) > (HEAP32[DYNAMICTOP_PTR>>2]|0)) segfault();
   if ((bytes|0) == 4) {
     if ((dest&3)) alignfault();
     return HEAP32[dest>>2] | 0;
@@ -1116,7 +1119,7 @@ function SAFE_HEAP_LOAD_D(dest, bytes) {
   dest = dest | 0;
   bytes = bytes | 0;
   if ((dest|0) <= 0) segfault();
-  if ((dest + bytes|0) > (DYNAMICTOP|0)) segfault();
+  if ((dest + bytes|0) > (HEAP32[DYNAMICTOP_PTR>>2]|0)) segfault();
   if ((bytes|0) == 8) {
     if ((dest&7)) alignfault();
     return +HEAPF64[dest>>3];
@@ -1313,6 +1316,14 @@ def emscript_wasm_backend(infile, settings, outfile, libraries=None, compiler_en
     backend_args += ['-thread-model=single'] # no threads support in backend, tell llc to not emit atomics
     # disable slow and relatively unimportant optimization passes
     backend_args += ['-combiner-alias-analysis=false', '-combiner-global-alias-analysis=false']
+
+    # asm.js-style exception handling
+    if settings['DISABLE_EXCEPTION_CATCHING'] != 1:
+      backend_args += ['-enable-emscripten-cxx-exceptions']
+    if settings['DISABLE_EXCEPTION_CATCHING'] == 2:
+      whitelist = ','.join(settings['EXCEPTION_CATCHING_WHITELIST'] or ['__fake'])
+      backend_args += ['-emscripten-cxx-exceptions-whitelist=' + whitelist]
+
     if DEBUG:
       logging.debug('emscript: llvm wasm backend: ' + ' '.join(backend_args))
       t = time.time()
@@ -1389,16 +1400,21 @@ def emscript_wasm_backend(infile, settings, outfile, libraries=None, compiler_en
   for line in open(wasm).readlines():
     if line.startswith('  (import '):
       parts = line.split(' ')
-      metadata['declares'].append(parts[3][1:])
+      # Don't include Invoke wrapper names (for asm.js-style exception handling)
+      # in metadata[declares], the invoke wrappers will be generated in
+      # this script later.
+      func_name = parts[3][1:]
+      if not func_name.startswith('invoke_'):
+        metadata['declares'].append(func_name)
     elif line.startswith('  (func '):
       parts = line.split(' ')
-      func = parts[3][1:]
-      metadata['implementedFunctions'].append(func)
+      func_name = parts[3][1:]
+      metadata['implementedFunctions'].append(func_name)
     elif line.startswith('  (export '):
       parts = line.split(' ')
-      exportname = parts[3][1:-1]
-      assert asmjs_mangle(exportname) not in metadata['exports']
-      metadata['exports'].append(exportname)
+      export_name = parts[3][1:-1]
+      assert asmjs_mangle(export_name) not in metadata['exports']
+      metadata['exports'].append(export_name)
 
   metadata['declares'] = filter(lambda x: not x.startswith('emscripten_asm_const'), metadata['declares']) # we emit those ourselves
 
@@ -1513,7 +1529,8 @@ return ASM_CONSTS[code](%s);
   outfile.write(pre)
   pre = None
 
-  basic_funcs = ['abort', 'assert']
+  basic_funcs = ['abort', 'assert', 'enlargeMemory', 'getTotalMemory']
+  if settings['ABORTING_MALLOC']: basic_funcs += ['abortOnCannotGrowMemory']
 
   access_quote = access_quoter(settings)
 
@@ -1530,9 +1547,21 @@ return ASM_CONSTS[code](%s);
   def math_fix(g):
     return g if not g.startswith('Math_') else g.split('_')[1]
 
-  basic_vars = ['STACKTOP', 'STACK_MAX', 'ABORT']
+  basic_vars = ['STACKTOP', 'STACK_MAX', 'DYNAMICTOP_PTR', 'ABORT']
   basic_float_vars = []
 
+  # Asm.js-style exception handling: invoke wrapper generation
+  invoke_wrappers = ''
+  with open(wasm) as f:
+    for line in f:
+      if line.startswith('  (import '):
+        parts = line.split(' ')
+        func_name = parts[3][1:]
+        if func_name.startswith('invoke_'):
+          sig = func_name[len('invoke_'):]
+          invoke_wrappers += '\n' + shared.JS.make_invoke(sig) + '\n'
+          basic_funcs.append(func_name)
+
   # sent data
   the_global = '{}'
   sending = '{ ' + ', '.join(['"' + math_fix(s) + '": ' + s for s in basic_funcs + global_funcs + basic_vars + basic_float_vars + global_vars]) + ' }'
@@ -1590,6 +1619,8 @@ Runtime.setTempRet0 = asm['setTempRet0'];
 Runtime.getTempRet0 = asm['getTempRet0'];
 ''')
 
+  funcs_js.append(invoke_wrappers)
+
   for i in range(len(funcs_js)): # do this loop carefully to save memory
     if WINDOWS: funcs_js[i] = funcs_js[i].replace('\r\n', '\n') # Normalize to UNIX line endings, otherwise writing to text file will duplicate \r\n to \r\r\n!
     outfile.write(funcs_js[i])
@@ -1729,3 +1760,4 @@ WARNING: You should normally never use this! Use emcc instead.
 
 if __name__ == '__main__':
   _main()
+  sys.exit(0)
